import os
import re
import logging

from jinja2 import Environment


logger = logging.getLogger(__name__)


class DistNotFound(Exception):
    pass


def _filter_pkg_dists(dists, pkg_name, pkg_ver):
    pkg_ver_str = re.sub('-', '_', pkg_ver)
    prefix = '{0}-{1}'.format(pkg_name, pkg_ver_str)
    return (d for d in dists if d.startswith(prefix))


def find_pkg_dists(project_path, dist_dir, pkg_name, pkg_ver):
    dist_dir = os.path.join(project_path, dist_dir)
    logger.info('Looking for package dists in {0}'.format(dist_dir))
    dists = _filter_pkg_dists(os.listdir(dist_dir), pkg_name, pkg_ver)
    dists = [{'pkg': pkg_name,
             'artifact': f,
              'path': os.path.join(dist_dir, f)}
             for f in dists]
    return dists


def build_index(title, items, index_type='root'):
    tmpl = """<!DOCTYPE html>
<html>
<head>
    <meta charset="UTF-8">
    <title>{{title}}</title>
</head>
<body>
    {%- if index_type != 'root' %}
    <h1>{{title}}</h1>
    {% endif -%}
    {% for item in items %}
    <a href="{{item}}">{{item}}</a><br>
    {% endfor %}
</body>
</html>
"""
    env = Environment()
    template = env.from_string(tmpl)
    return template.render(title=title, items=items,
                           index_type=index_type)


def is_dist_published(storage, dist):
    path = storage.join_path(dist['pkg'], dist['artifact'])
    logger.info('Ensuring dist is not already published: {0}'.format(path))
    return storage.path_exists(path)


def upload_dist(storage, dist):
    logger.info('Uploading dist: {0}'.format(dist['artifact']))
    dest = storage.join_path(dist['pkg'], dist['artifact'])
    storage.put_file(dist['path'], dest, sync=True)


def update_pkg_index(storage, pkg_name):
    logger.info('Updating index for package: {0}'.format(pkg_name))
    dists = [d for d in storage.listdir(pkg_name) if d != 'index.html']
    title = 'Links for {0}'.format(pkg_name)
    index = build_index(title, dists, 'pkg')
    index_path = storage.join_path(pkg_name, 'index.html')
    storage.put_contents(index, index_path)


def update_root_index(storage):
    logger.info('Updating repository index')
    pkgs = sorted([p for p in storage.listdir('.') if p != 'index.html'])
    title = 'Private Index'
    index = build_index(title, pkgs, 'root')
    index_path = storage.join_path('index.html')
    storage.put_contents(index, index_path)


def publish_package(name, version, storage, project_path, dist_dir):
    dists = find_pkg_dists(project_path, dist_dir, name, version)
    if not dists:
        raise DistNotFound((
            'No package distribution found in path {0}'
        ).format(dist_dir))
    rebuild_index = False
    for dist in dists:
        if not is_dist_published(storage, dist):
            logger.info('Trying to publish dist: {0}'.format(dist['artifact']))
            upload_dist(storage, dist)
            rebuild_index = True
        else:
            logger.debug((
                'Dist already published: {0} [skipping]'
            ).format(dist['artifact']))
    if rebuild_index:
        logger.info('Updating index')
        update_pkg_index(storage, name)
        update_root_index(storage)
    else:
        logger.debug('No index update required as no new dists uploaded')



import os
from ConfigParser import SafeConfigParser


class Config(object):

    def __init__(self, path, env):
        self.path = os.path.expanduser(path)
        self.env = env
        self.c = SafeConfigParser()
        with open(self.path) as f:
            self.c.readfp(f)

    @property
    def storage(self):
        return self.c.get('storage', 'type')

    @property
    def storage_config(self):
        return dict(self.c.items(self.storage))



import os
import argparse
import logging

from .config import Config
from .storage import load_storage
from .publish import publish_package


logger = logging.getLogger(__name__)


LOGGING_FORMAT = '[%(asctime)s] %(levelname)s %(name)s: %(message)s'


def log_level(verbosity):
    if verbosity == 1:
        return logging.INFO
    if verbosity > 1:
        return logging.DEBUG
    return logging.WARN


def cmd_publish(args):
    config = Config(args.conf_path, os.environ)
    storage = load_storage(config)
    return publish_package(args.pkg_name,
                           args.pkg_ver,
                           storage,
                           args.project_path,
                           args.dist_dir)


def main():
    parser = argparse.ArgumentParser(description=(
        'Script for publishing python package on private pypi'
    ))
    parser.add_argument('-p', '--project-path', default='.',
                        help='Path to project [Default: current dir]')
    parser.add_argument('-c', '--conf-path', default='~/.pypi-private.cfg',
                        help='Path to config [Default: ~/.pypi-private.cfg]')
    parser.add_argument('-v', '--verbose', default=1, action='count')

    subparsers = parser.add_subparsers(help='subcommand help')

    publish = subparsers.add_parser('publish', help='Publish package')
    publish.add_argument('-d', '--dist-dir', default='dist',
                         help='Directory to look for built distributions')
    publish.add_argument('pkg_name')
    publish.add_argument('pkg_ver')
    publish.set_defaults(func=cmd_publish)

    args = parser.parse_args()

    logging.basicConfig(format=LOGGING_FORMAT)

    logging.getLogger('pypiprivate').setLevel(log_level(args.verbose))

    args.func(args)

    return 0



import re
import ast

from setuptools import setup


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('pypiprivate/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(f.read().decode('utf-8')).group(1)))


with open('./README.rst') as f:
    long_desc = f.read()


setup(
    name='pypiprivate',
    version=version,
    author='Vineet Naik',
    author_email='vineet@helpshift.com',
    url='https://github.com/helpshift/pypiprivate',
    license='MIT License',
    description='Private package management tool for Python projects',
    long_description=long_desc,
    install_requires=['Jinja2==2.10.0',
                      'boto3==1.5.27'],
    packages=['pypiprivate'],
    entry_points={
        'console_scripts': [
            'pypi-private = pypiprivate.cli:main'
        ]
    },
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.6',
    ]
)



import os
import errno
import shutil
import mimetypes
import logging

import boto3
from botocore.exceptions import ClientError


logger = logging.getLogger(__name__)


class StorageException(Exception):
    pass


class PathNotFound(StorageException):
    pass


class Storage(object):

    def join_path(self, *args):
        raise NotImplementedError

    def listdir(self, path):
        raise NotImplementedError

    def path_exists(self, path):
        raise NotImplementedError

    def put_contents(self, contents, dest, sync=False):
        raise NotImplementedError

    def put_file(self, src, dest, sync=False):
        raise NotImplementedError


class LocalFileSystemStorage(Storage):

    def __init__(self, base_path):
        self.base_path = base_path

    @classmethod
    def from_config(cls, config):
        storage_config = config.storage_config
        return cls(storage_config['base_path'])

    def join_path(self, *args):
        return os.path.join(*args)

    def listdir(self, path):
        path = self.join_path(self.base_path, path)
        try:
            return os.listdir(path)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise PathNotFound('Path {0} not found'.format(path))
            raise e

    def path_exists(self, path):
        path = self.join_path(self.base_path, path)
        return os.path.exists(path)

    def ensure_dir(self, path):
        if not os.path.exists(path):
            os.makedirs(path)

    def put_contents(self, contents, dest, sync=False):
        dest_path = self.join_path(self.base_path, dest)
        self.ensure_dir(os.path.dirname(dest_path))
        with open(dest_path, 'w') as f:
            f.write(contents)
        # In LocalFileSystemStorage sync makes no sense
        return dest_path

    def put_file(self, src, dest, sync=False):
        dest_path = self.join_path(self.base_path, dest)
        self.ensure_dir(os.path.dirname(dest_path))
        shutil.copy(src, dest_path)
        return dest_path

    def __repr__(self):
        return (
            '<LocalFileSystemStorage(base_path="{0}")>'
        ).format(self.base_path)


class AWSS3Storage(Storage):

    def __init__(self, bucket, creds, acl, prefix=None):
        access_key, secret_key = creds
        session = boto3.Session(aws_access_key_id=access_key,
                                aws_secret_access_key=secret_key)
        self.s3 = s3 = session.resource('s3')
        self.bucket = s3.Bucket(bucket)
        self.prefix = prefix
        self.acl = acl

    @classmethod
    def from_config(cls, config):
        storage_config = config.storage_config
        env = config.env
        bucket = storage_config['bucket']
        prefix = storage_config.get('prefix')
        acl = storage_config.get('acl', 'private')
        creds = (env['PP_S3_ACCESS_KEY'], env['PP_S3_SECRET_KEY'])
        return cls(bucket, creds, acl, prefix)

    def join_path(self, *args):
        return '/'.join(args)

    def prefixed_path(self, path):
        parts = []
        if self.prefix:
            parts.append(self.prefix)
        if path != '.':
            parts.append(path)
        return self.join_path(*parts)

    def listdir(self, path):
        path = self.prefixed_path(path)
        if path != '' and not path.endswith('/'):
            s3_prefix = '{0}/'.format(path)
        else:
            s3_prefix = path
        logger.debug('Listing objects prefixed with: {0}'.format(s3_prefix))
        client = self.s3.meta.client
        paginator = client.get_paginator('list_objects')
        response = paginator.paginate(Bucket=self.bucket.name,
                                      Prefix=s3_prefix,
                                      Delimiter='/')
        file_objs = [c for c in response.search('Contents') if c]
        dir_objs = [cp for cp in response.search('CommonPrefixes') if cp]
        # If no objs found, it means the path doesn't exist
        if len(file_objs) == len(dir_objs) == 0:
            raise PathNotFound('Path {0} not found'.format(s3_prefix))
        files = (c['Key'][len(s3_prefix):] for c in file_objs)
        files = [f for f in files if f != '']
        dirs = [cp['Prefix'][len(s3_prefix):].rstrip('/') for cp in dir_objs]
        return files + dirs

    def path_exists(self, path):
        path = self.prefixed_path(path)
        logger.debug('Checking if key exists: {0}'.format(path))
        client = self.s3.meta.client
        try:
            client.head_object(Bucket=self.bucket.name, Key=path)
        except ClientError as e:
            logger.debug('Handled ClientError: {0}'.format(e))
            return False
        else:
            return True

    @staticmethod
    def _guess_content_type(path, default='application/octet-stream'):
        ctype = mimetypes.guess_type(path)[0] or default
        logger.debug('Guessed ctype of "{0}": "{1}"'.format(path, ctype))
        return ctype

    def put_contents(self, contents, dest, sync=False):
        dest_path = self.prefixed_path(dest)
        client = self.s3.meta.client
        logger.debug('Writing content to s3: {0}'.format(dest_path))
        client.put_object(Bucket=self.bucket.name,
                          Key=dest_path,
                          Body=contents.encode('utf-8'),
                          ContentType=self._guess_content_type(dest),
                          ACL=self.acl)
        if sync:
            waiter = client.get_waiter('object_exists')
            waiter.wait(Bucket=self.bucket.name, Key=dest_path)

    def put_file(self, src, dest, sync=False):
        dest_path = self.prefixed_path(dest)
        client = self.s3.meta.client
        logger.debug('Uploading file to s3: {0} -> {1}'.format(src, dest_path))
        with open(src) as f:
            client.put_object(Bucket=self.bucket.name,
                              Key=dest_path,
                              Body=f,
                              ContentType=self._guess_content_type(dest),
                              ACL=self.acl)
        if sync:
            waiter = client.get_waiter('object_exists')
            waiter.wait(Bucket=self.bucket.name, Key=dest_path)

    def __repr__(self):
        return (
            '<AWSS3Storage(bucket="{0}", prefix="{1}")>'
        ).format(self.bucket.name, self.prefix)


def load_storage(config):
    if config.storage == 'local-filesystem':
        return LocalFileSystemStorage.from_config(config)
    elif config.storage == 'aws-s3':
        return AWSS3Storage.from_config(config)
    else:
        raise ValueError('Unsupported storage "{0}"'.format(config.storage))
